/*
 * afcrypto.cpp:
 *
 * command for dealing with encryption issues
 */

/*
 * Copyright (c) 2007
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */

#include "affconfig.h"
#include "afflib.h"
#include "afflib_i.h"

#include <stdio.h>
#include <algorithm>
#include <vector>

char *progname = "afcrypto";

void change_passphrase(const char *fn,const char *old_passphrase,const char *new_passphrase)
{
    int fail = 0;

    AFFILE *af = af_open(fn,O_RDWR,0666);
    if(!af) err(1,fn);
    if(af_change_aes_passphrase(af,old_passphrase,new_passphrase)){
	warnx("%s: af_change_aes_passphrase failed",fn);
	fail = 1;
    }
    af_close(af);
    if(!fail) printf("%s: passphrase changed.\n",fn);
}

void get_and_change_passphrase(const char *fn)
{
    char old_passphrase[1024];
    char new_passphrase[1024];

    memset(old_passphrase,0,sizeof(old_passphrase));
    memset(new_passphrase,0,sizeof(new_passphrase));

    printf("Enter old passphrase: ");
    fgets(old_passphrase,sizeof(old_passphrase),stdin);
    char *cc = strchr(old_passphrase,'\n');if(cc) *cc='\000';

    /* See if this passphrase works*/

    AFFILE *af = af_open(fn,O_RDONLY,0666);
    if(!af) err(1,fn);
    if(af_use_aes_passphrase(af,old_passphrase)){
	errx(1,"passphrase incorrect");
    }
    af_close(af);

    printf("Enter new passphrase: ");
    fgets(new_passphrase,sizeof(new_passphrase),stdin);
    cc = strchr(new_passphrase,'\n');if(cc) *cc='\000';
    change_passphrase(fn,old_passphrase,new_passphrase);
}

void usage()
{
    printf("afcrypto version %s\n",PACKAGE_VERSION);
    printf("usage: afcrypto [options] filename.aff [filename2.aff ... ]\n");
    printf("   prints if each file is encrypted or not.\n");
    printf("options:\n");
    printf("    -c      --- change passphrase (take old and new from stdin)\n");
    printf("    -O old  --- specify old passphrase\n");
    printf("    -N new  --- specify old passphrase\n");
    printf("    -p passphrase --- checks to see if passphrase is the passphrase of the file\n");
    printf("                exit code is 0 if it is, -1 if it is not\n");
    printf("    -r file --- Open the file and read each line to see if it is the passphrase.\n");
    printf("    -V      --- Just print the version number and exit.\n");
    printf("Note: This program ignores the environment variables:\n");
    puts(AFFLIB_PASSPHRASE);
    puts(AFFLIB_PASSPHRASE_FILE);
    puts(AFFLIB_PASSPHRASE_FD);
    exit(0);
}

int check_file(AFFILE *af,const char *passphrase_file)
{
    int ret = -1;
    FILE *f = fopen(passphrase_file,"r");
    char buf[1024];
    memset(buf,0,sizeof(buf));
    while(!feof(f) && fgets(buf,sizeof(buf)-1,f)){
	char *cc = index(buf,'\n');
	if(cc) *cc = 0;
	if(af_use_aes_passphrase(af,buf)==0){
	    puts(buf);
	    ret = 0;
	}
    }
    fclose(f);
    return ret;
}

int main(int argc,char **argv)
{
    int bflag, ch;
    const char *old_passphrase=0;
    const char *new_passphrase=0;
    const char *check_passphrase = 0;
    const char *passphrase_file = 0;
    const char *progname = argv[0];
    int opt_quiet=0;

    /* Don't use auto-supplied passphrases */
    unsetenv(AFFLIB_PASSPHRASE);
    unsetenv(AFFLIB_PASSPHRASE_FILE);
    unsetenv(AFFLIB_PASSPHRASE_FD);

    bflag = 0;
    int opt_change = 0;
    while ((ch = getopt(argc, argv, "ch?VO:N:p:r:q")) != -1) {
	switch (ch) {
	case 'q': opt_quiet = 1;break;
	case 'c': opt_change = 1; break;
	case 'O': old_passphrase = optarg;break;
	case 'N': new_passphrase = optarg;break;
	case 'p': check_passphrase = optarg;break;
	case 'r': passphrase_file = optarg;break;
	case 'h':
	case '?':
	default:
	    usage();
	    break;
	case 'V':
	    printf("%s version %s\n",progname,PACKAGE_VERSION);
	    exit(0);
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }

    while(argc--){
	const char *fname = *argv++;
	AFFILE *af = af_open(fname,O_RDONLY,0);
	if(!opt_quiet){
	    printf("%s: %s encrypted segments\n",fname,af_has_encrypted_segments(af) ? "has" : "no");
	}
	if(check_passphrase){
	    if(af_use_aes_passphrase(af,check_passphrase)==0){
		printf("passphrase is correct\n");
	    }
	    else {
		printf("passphrase incorrect\n");
	    }
	    af_use_aes_passphrase(af,0);
	}
	if(passphrase_file){
	    if(check_file(af,passphrase_file)){
		fprintf(stderr,"%s: no valid passphrase in file\n",progname);
	    }
	}
	af_close(af);
	if(opt_change){
	    if(old_passphrase && new_passphrase) change_passphrase(argv[0],old_passphrase,new_passphrase);
	    else get_and_change_passphrase(argv[0]);
	}
	
    }
    return(0);
}
